/*
 * Autogenerated by the Meson build system.
 * Do not edit, your changes will be lost.
 */

#pragma once

#define NPUPNP_VERSION_MAJOR 6

#define NPUPNP_VERSION_MINOR 2

#define NPUPNP_VERSION_PATCH 1

#define NPUPNP_VERSION_STRING "6.2.1"

#define UPNP_ENABLE_IPV6

#undef UPNP_ENABLE_UNSPECIFIED_SERVER

#define UPNP_HAVE_CLIENT 1

#define UPNP_HAVE_DEBUG 0

#define UPNP_HAVE_DEVICE 1

#define UPNP_HAVE_GENA 1

#define UPNP_HAVE_OPTSSDP

#define UPNP_HAVE_SOAP 1

#define UPNP_HAVE_SSDP 1

#define UPNP_HAVE_TOOLS 1

#define UPNP_HAVE_WEBSERVER 1

#define USE_EXPAT

